/*
 * Tool name   : TCPSniffer
 * Description : A tool to sniff on TCP connections on a 
 *               Windows system and extract data content. 
 * Author      : Ruben Unteregger
 * Webpage     : http://www.megapanzer.com
 * Version     : 0.2
 * OS          : Tested on Microsoft Windows XP
 * Todo        : 
 *
 * Changes     : src/dst ip filter added
 * 
 *
 *
 *
 * License     :
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <stdio.h>     
#include <winsock2.h>
#include <Shlwapi.h>
#include "TCPSniffer.h"


#pragma comment(lib, "ws2_32.lib")
#pragma comment(lib, "Shlwapi.lib")



TCP_PATTERN gTCPPatterns[128];
DWORD gPatternCounter = 0;





/*
 * Entry point
 *
 */

int main()
{
  int lRetVal = 0;
  int lCounter = 0;
  char lTempBuf[65536];
  int  lBytesReceived;
  SOCKET lSocket;
  struct in_addr lAddr;
  int lSniffInterface;	
  char lHostName[MAX_BUF_SIZE + 1];
  struct hostent *lLocalEntry;
  WSADATA lWSA;	
  IPV4_HDR *lIPHeader = NULL;
  struct sockaddr_in lSrc;
  struct sockaddr_in lDst;



  /*
   * Initialise Winsock
   *
   */

  if (WSAStartup(MAKEWORD(2,2), &lWSA) != 0)
  {
    fprintf(stderr, "WSAStartup() failed : %d\n", GetLastError());
    lRetVal = 1;
    goto END;
  }
		
 
  /*
   * Create Raw Winsock
   *
   */
	
  if ((lSocket = socket(AF_INET, SOCK_RAW, IPPROTO_IP)) == INVALID_SOCKET)
  {
    fprintf(stderr, "Unable to create raw winsocket : %d\n", GetLastError());
    lRetVal = 2;
    goto END;
  }

		

  /*
   * Retrive the local hostname
   *
   */

	if (gethostname(lHostName, sizeof(lHostName) - 1) == SOCKET_ERROR) 
	{
        printf("Error : %d", WSAGetLastError());
		lRetVal = 3;
		goto END;
    }
	

  /*
   * Retrive all local IP addresses
   *
   */

	if ((lLocalEntry = gethostbyname(lHostName)) == NULL) 
	{
        printf("Error : %d.\n", WSAGetLastError());
 		lRetVal = 4;
		goto END;
    }


    for (lCounter = 0; lLocalEntry->h_addr_list[lCounter] != 0; ++lCounter) 
	{
		CopyMemory(&lAddr, lLocalEntry->h_addr_list[lCounter], sizeof(struct in_addr));
        printf("interface no. %d : %s\n", lCounter, inet_ntoa(lAddr));
    }

    ZeroMemory(&lSniffInterface, sizeof(lSniffInterface));
	printf("\n\nAttach sniffer to interface no. : ");
	scanf("%d", &lSniffInterface);
	


	ZeroMemory(&lDst, sizeof(lDst));
	CopyMemory(&lDst.sin_addr.s_addr, lLocalEntry->h_addr_list[lSniffInterface],sizeof(lDst.sin_addr.s_addr));
	lDst.sin_family = AF_INET;
	lDst.sin_port = 0;
    
	if (bind(lSocket, (struct sockaddr *) &lDst, sizeof(lDst)) == SOCKET_ERROR)
	{
		printf("Can't bind socket : %d\n", inet_ntoa(lAddr), GetLastError());
		lRetVal = 5;
		goto END;
	}
		

	//Enable this socket with the power to sniff : SIO_RCVALL is the key Receive ALL ;)
	
	lCounter = 1;

	if (WSAIoctl(lSocket, SIO_RCVALL, &lCounter, sizeof(lCounter), 0, 0, &lSniffInterface, 0, 0) == SOCKET_ERROR)
	{
		printf("WSAIoctl() failed.\n");
		lRetVal = 6;
		goto END;
	}
	

	// Initialize patterns
	initPatterns();

	/*
	 * Sniffer looop
	 *
	 */

	do
	{
		if((lBytesReceived = recvfrom(lSocket, lTempBuf, 65536, 0, 0, 0)) > 0)
		{
            lIPHeader = (IPV4_HDR *) lTempBuf;
            if (lIPHeader->ip_protocol == 6)
			{
				filterPacket(lTempBuf, lBytesReceived);
			}
		} else {
			printf( "recvfrom() failed : %d\n", GetLastError());
		}
	} while (lBytesReceived > 0);

END:	


	//End
	closesocket(lSocket);
	WSACleanup();

	return(lRetVal);
}







/*
 *
 *
 */

int initPatterns()
{
  int lRetVal = 0;
  FILE *lFH = NULL;
  char lTemp[MAX_BUF_SIZE + 1];
  char *lTempPtr = NULL;
  char lSrcPort[MAX_BUF_SIZE + 1];
  char lDstPort[MAX_BUF_SIZE + 1];


  ZeroMemory(&gTCPPatterns, sizeof(gTCPPatterns));
  gPatternCounter = 0;


  if ((lFH = fopen(PATTERN_FILE, "r")) != NULL)
  {
    ZeroMemory(lTemp, sizeof(lTemp));

    while (fgets(lTemp, sizeof(lTemp) - 1, lFH) != NULL)
    {

      /*
	   * Remove trailing newlines
	   *
	   */

      while (lTemp[strlen(lTemp) - 1] == '\r' || lTemp[strlen(lTemp) - 1] == '\n')
        lTemp[strlen(lTemp) - 1] = 0;


      if (lTemp[0] != '#')
	  {
        if (charCount(':', lTemp, strnlen(lTemp, sizeof(lTemp) - 1)) == 4)
	    {
          lTempPtr = lTemp;
          ZeroMemory(lSrcPort, sizeof(lSrcPort));
          ZeroMemory(lDstPort, sizeof(lDstPort));

          // Source address
		  strncpy(gTCPPatterns[gPatternCounter].sSourceAddr, lTempPtr, strchr(lTempPtr, ':') - lTempPtr);
          lTempPtr = strchr(lTempPtr, ':') + 1;

          // Source port
		  strncpy(lSrcPort, lTempPtr, strchr(lTempPtr, ':') - lTempPtr);
		  gTCPPatterns[gPatternCounter].sSourcePort = atoi(lSrcPort);
          lTempPtr = strchr(lTempPtr, ':') + 1;

          // Destination addrerss
		  strncpy(gTCPPatterns[gPatternCounter].sDestAddr, lTempPtr, strchr(lTempPtr, ':') - lTempPtr);
          lTempPtr = strchr(lTempPtr, ':') + 1;

          // Destination port
		  strncpy(lDstPort, lTempPtr, strchr(lTempPtr, ':') - lTempPtr);
		  gTCPPatterns[gPatternCounter].sDestPort = atoi(lDstPort);
          lTempPtr = strchr(lTempPtr, ':') + 1;

          // Patterns
		  strncpy(gTCPPatterns[gPatternCounter].sPattern, lTempPtr, strlen(lTempPtr));
          gPatternCounter++;
        } // if (charC...
	  } // if (lTemp[0]...

      ZeroMemory(lTemp, sizeof(lTemp));
    } // while (fgets(l...

    fclose(lFH);
  }

  return(lRetVal);
}



/*
 *
 *
 */

int charCount(char pSearchChar, char *pDataString, int pStrLen)
{
  int lRetVal = 0;
  int lCounter = 0;

  for (;lCounter < pStrLen; lCounter++)
    if (pDataString[lCounter] == pSearchChar)
      lRetVal++;


  return(lRetVal);
}






/*
 *
 *
 */

int filterPacket(unsigned char *pTempBuf, int pBytesReceived)
{
  int lRetVal = 0;
  IPV4_HDR *lIPHeader = (IPV4_HDR *) pTempBuf;
  int lIPHeaderLen = lIPHeader->ip_header_len * 4;
  TCP_HDR *lTCPHeader = (TCP_HDR *) (pTempBuf + lIPHeaderLen);
  char *lTCPData = NULL;
  int lTCPDataLen = 0;
  DWORD lCounter = 0;
  char lTemp[MAX_BUF_SIZE + 1];
  char lSrcAddr[64];
  char lDstAddr[64];
  struct sockaddr_in lSrc;
  struct sockaddr_in lDst;

  /*
   * Prepare IP information
   *
   */

  ZeroMemory(lSrcAddr, MAX_BUF_SIZE + 1);
  ZeroMemory(lDstAddr, MAX_BUF_SIZE + 1);

  lSrc.sin_addr.s_addr = lIPHeader->ip_srcaddr;
  lDst.sin_addr.s_addr = lIPHeader->ip_destaddr;

  strncpy (lSrcAddr, inet_ntoa(lSrc.sin_addr), 16); 
  strncpy (lDstAddr, inet_ntoa(lDst.sin_addr), 16); 


  /*
   * Prepare TCP header data
   *
   */

  lTCPData = pTempBuf + lIPHeaderLen + lTCPHeader->data_offset * 4;
  lTCPDataLen = pBytesReceived - lTCPHeader->data_offset * 4 - lIPHeader->ip_header_len * 4;


  /*
   * Match and find!
   *
   */


  if (lTCPDataLen > 0)
  {
    for (lCounter = 0; lCounter < gPatternCounter; lCounter++)
	{

      /*
	   * Check destination address/port
	   *
	   */


      if (strlen(gTCPPatterns[lCounter].sDestAddr) > 0 &&
          strncmp(lDstAddr, gTCPPatterns[lCounter].sDestAddr, 64) != 0)
      {
        continue;
      }


      if (gTCPPatterns[lCounter].sDestPort > 0 && 
		  gTCPPatterns[lCounter].sDestPort != ntohs(lTCPHeader->dest_port))
	  {
        continue;
      }

      /*
	   * Check source address/port
	   *
	   */

      if (strlen(gTCPPatterns[lCounter].sSourceAddr) > 0 &&
          strncmp(lSrcAddr, gTCPPatterns[lCounter].sSourceAddr, 64) != 0)
      {
        continue;
      }

	  if (gTCPPatterns[lCounter].sSourcePort > 0 && 
		  gTCPPatterns[lCounter].sSourcePort != ntohs(lTCPHeader->source_port))
	  {
        continue;
      }



      /*
	   * Check pattern
	   *
	   */

	  if (strlen(gTCPPatterns[lCounter].sPattern) != 0 && 
		  StrStrI(lTCPData, gTCPPatterns[lCounter].sPattern) == NULL)
	  {
        continue;
      }

      ZeroMemory(lTemp, sizeof(lTemp));
	  if (strlen(gTCPPatterns[lCounter].sPattern) > 0)
        strncpy(lTemp, StrStrI(lTCPData, gTCPPatterns[lCounter].sPattern), 50);
	  else
        strncpy(lTemp, lTCPData, 50);



	  printf("\n%s:%d -> %s:%d  - \"%s\"\n", lSrcAddr, 
                                                  ntohs(lTCPHeader->source_port),
												  lDstAddr,
												  ntohs(lTCPHeader->dest_port),
												  gTCPPatterns[lCounter].sPattern);


	  printf("\t%s\r\n...\r\n", lTemp);
    }
  }

  return(lRetVal);
}